# coding: utf-8
from __future__ import unicode_literals
from builtins import object
from gqlspection import log
import gqlspection


class GQLTypeProxy(object):
    name = ''
    schema = None
    _upstream = None
    max_depth = 4

    def __init__(self, name, schema):
        self.name = name
        self.schema = schema

    @property
    def upstream(self):
        # use cached value if present
        if self._upstream:
            return self._upstream

        if self.name in self.schema.types:
            self._upstream = self.schema.types[self.name]
            return self._upstream
        else:
            if log.is_debug:
                import pdb
                pdb.set_trace()
                log.debug("Found an unknown type: '%s'. At this time following types are present in schema:", self.name)
                for t in self.schema.types:
                    log.debug("    %s(%s) [%s]" % (type(t.name), t.name, t.kind.kind))

            raise Exception("GQLTypeProxy: type '%s' not defined" % self.name)

    def _proxy_getattr(self, item, levels):
        if levels >= self.max_depth:
            raise Exception("GQLTypeProxy: reached the recursion limit!")
        return getattr(self, item)

    def __getattr__(self, item):
        proxy = getattr(self.upstream, '_proxy_getattr', None)
        if proxy:
            # nested object detected, pass execution to proxy
            return proxy(item, 0)

        return getattr(self.upstream, item)

    def __dir__(self):
        return super(gqlspection.GQLType, self.upstream).__dir__()



# coding: utf-8
from __future__ import unicode_literals
from builtins import object


class GQLTypeKind(object):
    """Construct a type out of GraphQL schema fragment.

    Type kinds are stored in the 'kind' property, which is defined by __TypeKind enum:

        enum __TypeKind {
          SCALAR
          OBJECT
          INTERFACE
          UNION
          ENUM
          INPUT_OBJECT
          LIST             # wrapping type, need to check 'ofType' to construct the whole type
          NON_NULL         # wrapping type, need to check 'ofType' to construct the whole type
        }

    Additionally, note that the introspection_query we use only queries first level data, the nested types only return
    name and kind. That is, in order to query fields, interfaces, possibleTypes, enumValues, inputFields, inputFields,
    ofType and specifiedByURL you need to use the first-level definition within the schema[].

    A typedef is schema fragment such as:

        {u'kind': u'NON_NULL',
         u'name': None,
         u'ofType': {u'kind': u'LIST',
          u'name': None,
          u'ofType': {u'kind': u'NON_NULL',
           u'name': None,
           u'ofType': {u'kind': u'OBJECT',
            u'name': u'BillingPlanV2',
            u'ofType': None}}}}

    """
    wrapping_types = ('LIST', 'NON_NULL')
    non_wrapping_types = ('SCALAR', 'OBJECT', 'INTERFACE', 'UNION', 'ENUM', 'INPUT_OBJECT')
    leaf_types = ('SCALAR', 'ENUM')
    builtin_scalars = ('Int', 'Float', 'String', 'Boolean', 'ID')

    # in the example above: [NON_NULL, LIST, NON_NULL]
    modifiers = None
    # in the example above: OBJECT
    kind = None
    # in the example above: 'BillingPlanV2'
    name = None

    def __init__(self, name, kind, modifiers=None):
        self.name = name
        self.kind = kind
        self.modifiers = modifiers or []

    @staticmethod
    def from_json(typedef):
        current = typedef
        modifiers = []
        while current['kind'] in GQLTypeKind.wrapping_types:
            # iterate through intermediate modifiers (LIST and NON_NULL)
            modifiers.append(current['kind'])
            current = current['ofType']

        # by this time all modifiers should have been parsed, make sure the result is what we expect
        if current['kind'] not in GQLTypeKind.non_wrapping_types:
            raise Exception("GQLTypeKind: Type '%s' is of unknown kind: '%s'" % (typedef['name'], typedef['kind']))

        return GQLTypeKind(
            name=current['name'],
            kind=current['kind'],
            modifiers=modifiers
        )

    # String representation (in the example above: "[BillingPlanV2!]!")
    def __repr__(self):
        string_representation = self.name
        for modifier in reversed(self.modifiers):
            if modifier == 'NON_NULL':
                string_representation += '!'
            if modifier == 'LIST':
                string_representation = "[%s]" % string_representation

        return string_representation

    @property
    def is_builtin_scalar(self):
        return self.name in self.builtin_scalars

    @property
    def is_leaf(self):
        return self.kind in self.leaf_types

    @property
    def is_final(self):
        return self.is_leaf or self.is_builtin_scalar



# coding: utf-8
from __future__ import unicode_literals
from builtins import str, object
import gqlspection


class GQLArg(object):
    name         = ''
    kind         = None
    description  = ''
    type         = None
    default_value = ''

    def __init__(self, name, kind, type_, description='', default_value=''):
        self.name = name
        self.kind = kind
        self.type = type_
        self.description = description
        self.default_value = default_value

    @staticmethod
    def from_json(json, schema):
        kind = gqlspection.GQLTypeKind.from_json(json['type'])

        return GQLArg(
            name=json['name'],
            kind=kind,
            type_=gqlspection.GQLTypeProxy(kind.name, schema),
            description=json.get('description', ''),
            default_value=json.get('default_value', '')
        )

    def __repr__(self):
        return '{name}: {type}'.format(
            name = self.name,
            type = str(self.kind)
        )



# coding: utf-8
from __future__ import unicode_literals
from builtins import str, object
import gqlspection
from gqlspection.utils import pad_string


class GQLSubQuery(object):
    field       = None
    name        = ''
    description = ''
    max_depth   = 4

    def __init__(self, field, max_depth=5):
        self.field        = field
        self.name         = field.name
        self.max_depth    = max_depth - 1

    def __repr__(self):
        self.str()

    def __str__(self):
        self.str()

    @staticmethod
    def _indent(indent):
        """Generates characters that should be used for Space and Newline, depending on request indentation level.

        If indent > 0, space is preserved and new lines are started with 'indent' number of spaces.
        If indent = 0, space is preserved, but new lines are removed
        If indent = None, both space and newlines are trimmed.
        """
        NEWLINE = '\n' if indent else ''
        PADDING = ' ' * indent if indent else ''
        SPACE   = ' '  if (indent is not None) else ''

        return SPACE, NEWLINE, PADDING

    def str(self, pad=4):
        """Generate a string representation.

        'pad' parameter defines number of space characters to use for indentation. Special values:
        'pad=0'    generates minimized query (oneliner without comments).
        'pad=None' generates super-optimized query where spaces are omitted as much as possible
        """
        # whitespace characters collapse when query gets minimized
        SPACE, NEWLINE, PADDING = self._indent(pad)

        arguments = (',' + SPACE).join([str(x) for x in self.field.args])

        first_line = ''.join((
            self.name,
            "({arguments})".format(arguments=arguments) if arguments else "",
            (SPACE + "{" + NEWLINE) if not self.field.type.kind.is_final else NEWLINE
        ))

        middle_lines = ''
        if self.max_depth:
            for field in self.field.type.fields:
                subquery = GQLSubQuery(field, max_depth=self.max_depth)
                middle_lines += NEWLINE.join(subquery.str(pad).splitlines()) + NEWLINE
        else:
            # Max recursion depth reached
            middle_lines = '!!! MAX RECURSION DEPTH REACHED !!!' + NEWLINE
        middle_lines = pad_string(middle_lines, pad)

        last_line = '}' + NEWLINE if not self.field.type.kind.is_final else ""

        if pad:
            if self.description:
                description_line = gqlspection.utils.format_comment(self.description) + NEWLINE
            else:
                description_line = ''
            result = description_line + first_line + middle_lines + last_line
        else:
            result = first_line + middle_lines + last_line

        return result



# coding: utf-8
from __future__ import unicode_literals
from builtins import object
import gqlspection

# FIXME: Add support for InputFields


class GQLField(object):
    name = ''
    description = ''
    kind  = None
    type  = None
    args = None
    is_deprecated = False
    deprecation_reason = ''
    schema = None

    def __init__(self, name, kind, schema, description='', args=None, is_deprecated=False, deprecation_reason=''):
        self.name = name
        self.kind = kind
        self.type = gqlspection.GQLTypeProxy(kind.name, schema)
        self.schema = schema
        self.description = description
        self.args = args or gqlspection.GQLArgs(schema, {})
        self.is_deprecated = is_deprecated
        self.deprecation_reason = deprecation_reason

    @staticmethod
    def _wrap_args(json, schema):
        args = []
        for i in (json.get('args', []) or []):
            args.append(gqlspection.GQLArg.from_json(i, schema))
        return args

    @staticmethod
    def from_json(field, schema):
        return GQLField(
            name=field['name'],
            kind=gqlspection.GQLTypeKind.from_json(field['type']),
            schema=schema,
            description=field.get('description', ''),
            args=GQLField._wrap_args(field, schema),
            is_deprecated=field.get('isDeprecated', False),
            deprecation_reason=field.get('deprecationReason', '')
        )



# coding: utf-8
from __future__ import unicode_literals
from builtins import str
try:
    from collections.abc import Mapping
except ImportError:
    from collections import Mapping  # type: ignore
from collections import OrderedDict


class GQLList(Mapping):
    """A very specific data structure for internal use. Acts as an iterable in certain instances and a dict in others.

    Characteristics:

      - Gets initialized from an iterable of some type, which should have a 'name' field
      - Upon iteration, will return original dictionaries, sorted by the 'name' field
      - Allows selecting elements both by index (gqllist[3]) and by the 'name' (gqllist['some-name'])
      - GQLList is meant for read-only data, so there is no way to add, update, delete elements
    """
    _elements = None

    def __init__(self, elements):
        sorted_elements = sorted(elements, key=lambda i: i.name)
        self._elements = OrderedDict(
            ((i.name, i) for i in sorted_elements)
        )

    def __getitem__(self, item):
        if isinstance(item, str):
            return self._elements[item]
        if type(item) == int:
            key = list(self._elements.keys())[item]
            return self._elements[key]
        raise Exception("GQLList: unknown type: %s", type(item))

    def __iter__(self):
        return (self._elements[el] for el in self._elements)

    def __str__(self):
        return '\n'.join((str(el) for el in self._elements))

    def __repr__(self):
        first_line = "GQLList[{inner_type}]".format(
            inner_type=str(type(self._elements[0]))
        )

        other_lines = ['    ' + repr(el) for el in self._elements]

        return '\n'.join([first_line] + other_lines)

    def __bool__(self):
        return not not len(self._elements)

    def __len__(self):
        return len(self._elements)

    def __contains__(self, item):
        if isinstance(item, str):
            return item in list(self._elements.keys())
        else:
            return item in list(self._elements.values())
